'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var invariant = _interopDefault(require('invariant'));

const keys = Object.keys,
      getOwnPropertyDescriptors = Object.getOwnPropertyDescriptors,
      defineProperty = Object.defineProperty;
invariant(getOwnPropertyDescriptors, `funcadelic.js requires Object.getOwnPropertyDescriptors. See https://github.com/cowboyd/funcadelic.js#compatibility`);
invariant("name" in Function.prototype && "name" in function x() {}, `funcadelic.js requires Function.name. See https://github.com/cowboyd/funcadelic.js#compatibility`);
const VERSION = 0;
let uniqueTag = 0;
function type(Class) {
  let name = Class.name;

  if (!name) {
    throw new Error('invalid typeclass name: ' + name);
  }

  let symbolName = `@@funcadelic-${VERSION}/${name}/${uniqueTag++}`;
  let symbol = Symbol[symbolName] ? Symbol[symbolName] : Symbol[symbolName] = Symbol(symbolName);

  Class.for = function _for(value) {
    let i = value[symbol];

    if (!i) {
      throw new Error(`No instance found on ${value} of typeclass ${name}`);
    }

    return i;
  };

  Class.instance = function (constructor, methods) {
    defineProperty(constructor.prototype, symbol, {
      value: methods,
      configurable: true,
      enumerable: false
    });
  };

  Class.symbolName = symbolName;
  Class.symbol = symbol;
  let properties = getOwnPropertyDescriptors(Class.prototype);
  keys(properties).filter(key => key != 'constructor').forEach(key => {
    Class.prototype[key] = Class.prototype[key].bind(Class.for);
  });
  return Class;
}

const Semigroup = type(class Semigroup {
  static get name() {
    return 'Semigroup';
  }

  append(left, right) {
    let _this = this(left),
        append = _this.append;

    return append(left, right);
  }

});
const append = Semigroup.prototype.append;

const Foldable = type(class Foldable {
  static get name() {
    return 'Foldable';
  }

  foldr(fn, initial, foldable) {
    let _this = this(foldable),
        foldr = _this.foldr;

    return foldr(fn, initial, foldable);
  }

  foldl(fn, initial, foldable) {
    let _this2 = this(foldable),
        foldl = _this2.foldl;

    return foldl(fn, initial, foldable);
  }

  size(foldable) {
    let _this3 = this(foldable),
        foldr = _this3.foldr;

    return foldr(len => len + 1, 0, foldable);
  }

});
const _Foldable$prototype = Foldable.prototype,
      foldr = _Foldable$prototype.foldr,
      foldl = _Foldable$prototype.foldl,
      size = _Foldable$prototype.size;

const Stable = Symbol('Stable');
function stable(fn) {
  switch (fn.length) {
    case 0:
      return thunk(fn);

    case 1:
      return stableOne(fn);

    default:
      throw new Error(`Cannot (yet) make functions with ${fn.length} stable`);
  }
}
function thunk(fn) {
  if (fn[Stable]) {
    return fn;
  }

  let evaluated = false;
  let result = undefined;

  function evaluate() {
    if (evaluated) {
      return result;
    } else {
      result = fn.call(this);
      evaluated = true;
      return result;
    }
  }

  evaluate[Stable] = true;
  return evaluate;
}
function stableOne(fn) {
  if (fn[Stable]) {
    return fn;
  }

  let cache = new WeakMap();

  function stabilizedOne(argument) {
    if (!cache.has(argument)) {
      cache.set(argument, fn(argument));
    }

    return cache.get(argument);
  }

  stabilizedOne[Stable] = true;
  return stabilizedOne;
}

const getPrototypeOf = Object.getPrototypeOf,
      getOwnPropertyDescriptors$1 = Object.getOwnPropertyDescriptors,
      getOwnPropertySymbols = Object.getOwnPropertySymbols;
Semigroup.instance(Object, {
  append(o1, o2) {
    let properties = assign({}, getOwnPropertyDescriptors$1(o1), getOwnPropertyDescriptors$1(o2));
    return Object.create(getPrototypeOf(o1), stableize(properties));
  }

});

function assign(target, a, b) {
  function copy(source) {
    let keys = allKeys(source);
    let totalKeys = keys.length;

    for (let j = 0; j < totalKeys; j++) {
      let key = keys[j];
      target[key] = source[key];
    }
  }

  copy(a);
  copy(b);
  return target;
}

function stableize(properties) {
  return foldl((descriptors, key) => {
    let descriptor = properties[key];

    if (!descriptor.get) {
      descriptors[key] = descriptor;
    } else {
      let getter = descriptor.get;
      let cached = stable(instance => getter.call(instance));

      descriptor.get = function () {
        return cached(this);
      };

      descriptors[key] = descriptor;
    }

    return descriptors;
  }, {}, allKeys(properties));
}

function allKeys(o) {
  return Object.keys(o).concat(getOwnPropertySymbols(o));
}

Semigroup.instance(Array, {
  append(a1, a2) {
    return a1.concat(a2);
  }

});

const Functor = type(class Functor {
  static get name() {
    return 'Functor';
  }

  map(fn, f) {
    let _this = this(f),
        map = _this.map;

    return map(fn, f);
  }

});
const map = Functor.prototype.map;

const Monoid = type(class Monoid extends Semigroup {
  static get name() {
    return 'Monoid';
  }

  reduce(M, values) {
    let empty = this(M.prototype).empty();
    return foldl((reduction, value) => append(reduction, value), empty, values);
  }

});

Monoid.create = function create(Definition) {
  class Monoidal extends Definition {
    constructor(value = Monoidal.prototype.empty()) {
      super();
      this.value = value;
    }

  }

  Monoidal.reduce = list => reduce(Monoidal, map(i => new Monoidal(i), list)).value;

  Semigroup.instance(Monoidal, {
    append(a, b) {
      return new Monoidal(Monoidal.prototype.append(a.value, b.value));
    }

  });
  Monoid.instance(Monoidal, {
    empty() {
      return new Monoidal();
    }

  });
  return Monoidal;
};

const reduce = Monoid.prototype.reduce;

Monoid.instance(Object, {
  empty() {
    return {};
  }

});

Monoid.instance(Array, {
  empty() {
    return [];
  }

});

const getPrototypeOf$1 = Object.getPrototypeOf,
      keys$1 = Object.keys;
Functor.instance(Object, {
  map(fn, object) {
    let descriptors = keys$1(object).reduce((descriptors, key) => {
      descriptors[key] = {
        configurable: true,
        enumerable: true,
        get: stable(() => fn(object[key], key))
      };
      return descriptors;
    }, {});
    return Object.create(getPrototypeOf$1(object), descriptors);
  }

});

Functor.instance(Array, {
  map(fn, array) {
    return array.map(fn);
  }

});

Functor.instance(Promise, {
  map(fn, promise) {
    return promise.then(fn);
  }

});

function curry(f) {
  return function curried(...t) {
    if (t.length === 0) return curried;
    if (t.length >= f.length) return f(...t);
    return curried.bind(this, ...t);
  };
}
const Applicative = type(class Applicative extends Functor {
  static get name() {
    return 'Applicative';
  }

  pure(Type, value) {
    return this(Type.prototype).pure(value);
  }

  apply(Type, fn, list) {
    let _this = this(Type.prototype),
        pure = _this.pure,
        applyOne = _this.applyOne;

    let initial = pure(curry(fn));
    return foldl((left, right) => applyOne(left, right), initial, list);
  }

  applyOne(left, right) {
    this(left).applyOne(right);
  }

});
const _Applicative$prototyp = Applicative.prototype,
      pure = _Applicative$prototyp.pure,
      apply = _Applicative$prototyp.apply,
      applyOne = _Applicative$prototyp.applyOne;

Applicative.instance(Promise, {
  pure(value) {
    return Promise.resolve(value);
  },

  applyOne(left, right) {
    return Promise.all([left, right]).then(([fn, value]) => fn(value));
  }

});

const Monad = type(class Monad extends Applicative {
  static get name() {
    return 'Monad';
  }

  flatMap(fn, m) {
    return this(m).flatMap(fn, m);
  }

});
const flatMap = Monad.prototype.flatMap;

Monad.instance(Array, {
  flatMap(fn, array) {
    return reduce(Array, map(fn, array));
  }

});

Monad.instance(Promise, {
  flatMap(fn, promise) {
    return promise.then(fn);
  }

});

Foldable.instance(Array, {
  foldr(fn, initial, array) {
    return array.reduceRight(fn, initial);
  },

  foldl(fn, initial, array) {
    return array.reduce(fn, initial);
  }

});

const keys$2 = Object.keys;
Foldable.instance(Object, {
  foldr(fn, initial, object) {
    return keys$2(object).reduceRight((memo, key) => fn(memo, {
      key,

      get value() {
        return object[key];
      }

    }), initial);
  },

  foldl(fn, initial, object) {
    return keys$2(object).reduce((memo, key) => fn(memo, {
      key,

      get value() {
        return object[key];
      }

    }), initial);
  }

});

const Filterable = type(class Filterable {
  static get name() {
    return 'Filterable';
  }

  filter(fn, f) {
    return this(f).filter(fn, f);
  }

});
const filter = Filterable.prototype.filter;

const create = Object.create,
      getPrototypeOf$2 = Object.getPrototypeOf;
Filterable.instance(Object, {
  filter(fn, object) {
    let instance = create(getPrototypeOf$2(object));
    return foldr((filtered, entry) => {
      if (fn(entry)) {
        return append(filtered, {
          get [entry.key]() {
            return entry.value;
          }

        });
      } else {
        return filtered;
      }
    }, instance, object);
  }

});

Filterable.instance(Array, {
  filter(fn, array) {
    return array.filter(fn);
  }

});

class Chain {
  map(fn) {
    return new Thunk(() => map(fn, this.valueOf()));
  }

  flatMap(fn) {
    return new Thunk(() => flatMap(fn, this.valueOf()));
  }

  filter(fn) {
    return new Thunk(() => filter(fn, this.valueOf()));
  }

  append(thing) {
    return new Thunk(() => append(this.valueOf(), thing));
  }

  tap(fn) {
    fn(this.valueOf());
    return this;
  }

}

class Value extends Chain {
  constructor(value) {
    super();
    this.value = value;
  }

  valueOf() {
    return this.value;
  }

}

class Thunk extends Chain {
  constructor(fn) {
    super();
    this.fn = stable(fn);
  }

  valueOf() {
    return this.fn();
  }

}

function chain(value) {
  return new Value(value);
}

exports.type = type;
exports.Semigroup = Semigroup;
exports.append = append;
exports.Monoid = Monoid;
exports.reduce = reduce;
exports.Functor = Functor;
exports.map = map;
exports.Applicative = Applicative;
exports.pure = pure;
exports.apply = apply;
exports.Monad = Monad;
exports.flatMap = flatMap;
exports.Foldable = Foldable;
exports.foldr = foldr;
exports.foldl = foldl;
exports.Filterable = Filterable;
exports.filter = filter;
exports.stable = stable;
exports.chain = chain;
//# sourceMappingURL=funcadelic.cjs.js.map
